@@ -17,87 +17,78 @@ from rlmusician.utils import convert_to_base
 
 class CounterpointEnv(gym.Env):
     """
-    An environment where agent sequentially composes melodic lines.
+    An environment where counterpoint line is composed given cantus firmus.
     """
 
     reward_range = (-np.inf, np.inf)
+    valid_durations = [1, 2, 4, 8]
 
     def __init__(
             self,
             piece: Piece,
-            observation_decay: float,
-            reward_for_dead_end: float,
             scoring_coefs: Dict[str, float],
-            scoring_fn_params: Dict[str, Dict[str, Any]]
+            scoring_fn_params: Dict[str, Dict[str, Any]],
+            reward_for_dead_end: float,
+            verbose: bool = False
     ):
         """
         Initialize instance.
 
         :param piece:
-            data structure representing musical piece
-        :param observation_decay:
-            coefficient of exponential decay for previous piano roll's
-            time steps
-        :param reward_for_dead_end:
-            reward for situations where there aren't any allowed actions, but
-            piece is not finished
+            data structure representing musical piece with florid counterpoint
         :param scoring_coefs:
             mapping from scoring function names to their weights in final score
         :param scoring_fn_params:
             mapping from scoring function names to their parameters
+        :param reward_for_dead_end:
+            reward for situations where there aren't any allowed actions, but
+            piece is not finished
+        :param verbose:
+            if it is set to `True`, breakdown of reward is printed at episode
+            end
         """
         self.piece = piece
-        self.observation_decay = observation_decay
-        self.reward_for_dead_end = reward_for_dead_end
         self.scoring_coefs = scoring_coefs
         self.scoring_fn_params = scoring_fn_params
+        self.reward_for_dead_end = reward_for_dead_end
+        self.verbose = verbose
 
-        self.verbose = False
-
-        n_actions = len(piece.all_movements) ** len(piece.lines)
+        n_actions = len(piece.all_movements) * len(self.valid_durations)
         self.action_space = gym.spaces.Discrete(n_actions)
-        self.action_to_movements = None
-        self.__set_action_to_movements()
+        self.action_to_line_continuation = None
+        self.__set_action_to_line_continuation()
 
-        roll_height = piece.highest_row_to_show - piece.lowest_row_to_show + 1
         self.observation_space = gym.spaces.Box(
             low=0,
-            high=1 / (1 - observation_decay),
-            shape=(roll_height,),
-            dtype=np.float64
+            high=1,
+            shape=self.piece.piano_roll.shape,
+            dtype=np.int32
         )
 
-    def __set_action_to_movements(self) -> None:
-        """Create mapping from action to line movements."""
+    def __set_action_to_line_continuation(self) -> None:
+        """Create mapping from action to a pair of movement and duration."""
         base = len(self.piece.all_movements)
-        offset = self.piece.max_skip
-        n_lines = len(self.piece.lines)
-        action_to_movements = {
-            i: [x - offset for x in convert_to_base(i, base, n_lines)]
-            for i in range(self.action_space.n)
+        required_len = 2
+        raw_mapping = {
+            action: (x for x in convert_to_base(action, base, required_len))
+            for action in range(self.action_space.n)
         }
-        self.action_to_movements = action_to_movements
+        offset = self.piece.max_skip
+        action_to_continuation = {
+            action: (movement_id - offset, 2 ** duration_id)
+            for action, (duration_id, movement_id) in raw_mapping.items()
+        }
+        self.action_to_line_continuation = action_to_continuation
 
     @property
     def valid_actions(self) -> List[int]:
         """Get actions that are valid at the current step."""
         valid_actions = [
             i for i in range(self.action_space.n)
-            if self.piece.check_movements(self.action_to_movements[i])
+            if self.piece.check_validity(*self.action_to_line_continuation[i])
         ]
         return valid_actions
 
-    @property
-    def __decay_coefs(self) -> np.ndarray:
-        """Get decay coefficients for passed piano roll's steps."""
-        decay_coefs = np.empty((self.piece.last_finished_measure + 1,))
-        decay_coefs[0] = 1
-        decay_coefs[1:] = self.observation_decay
-        decay_coefs = np.cumprod(decay_coefs)
-        decay_coefs = np.flip(decay_coefs)
-        decay_coefs = decay_coefs.reshape((1, -1))
-        return decay_coefs
-
     def step(self, action: int) -> Tuple[np.ndarray, float, bool, Dict]:
         """
         Run one step of the environment's dynamics.
@@ -112,25 +103,19 @@ class CounterpointEnv(gym.Env):
                     `step()` calls will return undefined results,
             - info: list of next allowed actions
         """
-        # Act.
-        movements = self.action_to_movements[action]
-        self.piece.add_measure(movements)
+        movement, duration = self.action_to_line_continuation[action]
+        self.piece.add_line_element(movement, duration)
 
-        # Compute `observation`.
-        right_end = self.piece.last_finished_measure + 1
-        ready_piano_roll = self.piece.piano_roll[:, :right_end]
-        observation = np.sum(self.__decay_coefs * ready_piano_roll, axis=1)
-
-        # Compute `info`.
+        observation = self.piece.piano_roll
         info = {'next_actions': self.valid_actions}
 
-        # Compute `done`.
-        finish = self.piece.last_finished_measure == self.piece.n_measures - 1
+        past_duration = self.piece.current_time_in_eighths
+        piece_duration = self.piece.total_duration_in_eighths
+        finished = past_duration == piece_duration
         no_more_actions = len(info['next_actions']) == 0
-        done = finish or no_more_actions
+        done = finished or no_more_actions
 
-        # Compute `reward`.
-        if finish:
+        if finished:
             reward = evaluate(
                 self.piece,
                 self.scoring_coefs,
@@ -152,7 +137,7 @@ class CounterpointEnv(gym.Env):
             initial observation
         """
         self.piece.reset()
-        initial_observation = self.piece.piano_roll[:, 0]
+        initial_observation = self.piece.piano_roll
         return initial_observation
 
     def render(self, mode='human'):  # pragma: no cover.
